#!/usr/bin/env python

# Copyright (C) 2002 Jrg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import threading
import Queue
import sys
import ao

import events
import pcm
import mp3file

from helper import debug

class bufferedao(threading.Thread):
    def __init__(self, device, queuesize, rate, dsp):
        self.device = device
        self.queuesize = queuesize
        self.rate = rate
        self.dsp = dsp
        self.ao = ao.AudioDevice(device, rate=rate, options={"dsp": self.dsp})
        self.queue = Queue.Queue(self.queuesize)
        self.done = 0
        threading.Thread.__init__(self)
        
        # we don't know what to do, if we are left alone, hence:
        self.setDaemon(1)
        
    def run(self):
        while not self.done:
            buff, bytes = self.queue.get(1)
            if buff!=0 and bytes!=0:
                self.ao.play(buff, bytes)

    def play(self, buff, bytes):
        self.queue.put((buff, bytes))

    def stop(self):
        self.done = 1
        while self.queue.full():
            self.queue.get(1)
        self.play(0,0)


class madplayer(threading.Thread):

    def __init__(self, player, rqueue, squeue, device, dsp, bufsize, crossfadingstart, crossfadingduration):
        self.player = player
        self.device = device
        self.dsp = dsp
        self.rate = 44100
        self.SIZE = 4806
        if bufsize>0:
            queuesize = 1024*bufsize/self.SIZE + 1
            self.ao = bufferedao("oss", queuesize=queuesize, rate=self.rate, dsp=dsp)
            self.ao.start()
            
            # calulate maximal length of pcm samples in queue in 1/1000th seconds
            self.queuelength = 1000/4.0*queuesize*self.SIZE/self.rate
        else:
            self.ao = ao.AudioDevice("oss", rate=self.rate, options={"dsp": self.dsp})
            self.queuelength = 0

        # queues for communication between madplayer and player
        self.rqueue = rqueue
        self.squeue = squeue
        self.done = 0

        # songs playing
        self.songs = []

        # crossfading parameters (all values in seconds and change/second, resp.)
        self.crossfadingduration = crossfadingduration
        self.crossfadingstart = crossfadingstart
        self.crossfadingratio = 0
        self.crossfadingrate = 1.0/self.rate/self.crossfadingduration

        # have we already requested a new song for crossfading?
        self.requestednextsong = 0

        threading.Thread.__init__(self)

        # we are only servants of higher powers
        self.setDaemon(1)

    def processevents(self):
        "check for incoming events and process them"
        if not self.rqueue.empty() or len(self.songs)==0:
            event = self.rqueue.get(1)
            if isinstance(event, events.nextsong):
                debug("madplayer: new song: %s" % event.song.path)

                # we want maximally 2 songs in queue
                if len(self.songs)==2:
                    del self.songs[0]

                self.songs.append(mp3file.mp3file(event.song.path,
                                                  self.rate,
                                                  "%s - %s" % (event.song.artist,
                                                               event.song.title),
                                                  event.song.length))

                debug(" (%d songs in queue)\n" % len(self.songs))

                # now that we have a new song, we can reset this
                self.requestednextsong = 0

            elif event is None:
                self.songs = []
            elif event=="QUIT":
                self.done = 1

    def play(self):
        "decode songs and mix them together"
        if len(self.songs)==1:
            song = self.songs[0]
            buff = song.read(self.SIZE)
            if len(buff)>0:
                self.ao.play(buff, len(buff))
            else:
                debug("madplayer: song ends: %s (0 songs in queue)\n" % self.songs[0].path)
                del self.songs[0]

            if (song.ttime-song.ptime)/1000<self.crossfadingstart \
               and not self.requestednextsong:
                self.squeue.put(events.requestnextsong(self.player))
                self.requestednextsong = 1

        elif len(self.songs)==2:
            buff1 = self.songs[0].read(self.SIZE)
            buff2 = self.songs[1].read(self.SIZE)

            if len(buff1)==len(buff2):
                # we only crossfade, if the samples from both songs
                # have the same length, because we don't want to
                # fill with zeroes. Should we?
                buff, self.crossfadingratio = pcm.mix(buff1, buff2,
                                                 self.crossfadingratio,
                                                 self.crossfadingrate)
                if self.crossfadingratio>=1:
                    self.crossfadingratio = 0
                    debug("madplayer: song ends: %s (1 song in queue)\n" % self.songs[0].path)
                    del self.songs[0]
            elif len(buff1)<len(buff2):
                buff = buff2
                self.crossfadingratio = 0
                debug("madplayer: song ends: %s (1 song in queue)\n" % self.songs[0].path)
                del self.songs[0]
            else:
                # a currently faded in song ends => delete it together 
                # with faded out song. We don't want a revival of 
                # the latter!
                buff = ""
                self.crossfadingratio = 0
                debug("madplayer: song ends: %s (1 song in queue)\n" % self.songs[0].path)
                debug("madplayer: song ends: %s (0 song in queue)\n" % self.songs[1].path)
                self.songs = []

            if len(buff)>0:
                self.ao.play(buff, len(buff))

    def updatestatus(self):
        "update player status"
        if len(self.songs)==0:
            self.player.playing = 0
        else:
            if self.crossfadingratio:
                self.player.title = "-> %s" % self.songs[-1].title
            else:
                self.player.title = self.songs[-1].title
            self.player.ptime = max(self.songs[-1].ptime-self.queuelength, 0)
            self.player.ttime = self.songs[-1].ttime
            self.player.playing = 1 # must come last...


    def run(self):
        while not self.done:
            self.processevents()
            self.play()
            self.updatestatus()

        if isinstance(self.ao, bufferedao):
            self.ao.stop()


"""Note: A player can request a song from the playlist via the
   event requestnextsong(self)
"""

class player:
    def __init__(self, device="oss", dsp="/dev/dsp", bufsize=0, crossfadingstart=5, crossfadingduration=10):
        self.playing = 0
        self.title = ""
        self.ptime = 0
        self.ttime = 0
        self.rqueue = Queue.Queue(-1)
        self.squeue = Queue.Queue(-1)

        self.mad = madplayer(self,
                             self.squeue, self.rqueue,
                             device,
                             dsp,
                             bufsize,
                             crossfadingstart,
                             crossfadingduration)
        self.mad.start()
        
        events.ec.subscribe(events.timer, self.timer)
        events.ec.subscribe(events.playerstart, self.playerstart)
        events.ec.subscribe(events.playerstop, self.playerstop)
        events.ec.subscribe(events.playerforward, self.playerforward)
        events.ec.subscribe(events.nextsong, self.nextsong)

    def timer(self, event):
        """player update, regularly called by pytone"""
        if not self.rqueue.empty():
            events.ec.notify(self.rqueue.get(1))

        events.ec.notify(events.playbackinfochanged(self, 
            self.playing, self.title, self.ptime, self.ttime))

    def nextsong(self, event):
        """play event.song next"""
        if event.player is self:
            if event.song:
                self.squeue.put(event)

    def playerstart(self, event):
        """start playing"""
        if event.player is self and not self.playing:
            events.ec.notify(events.requestnextsong(self))

    def playerstop(self, event):
        """stop playing"""
        if event.player is self:
            self.squeue.put(None)
            events.ec.notify(events.playbackinfochanged(self))

    def playerforward(self, event):
        """immediately play next song"""
        if event.player is self:
            events.ec.notify(events.requestnextsong(self))

    def playerquit(self):
        self.squeue.put("QUIT")

